@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2023 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2023 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -25,50 +25,34 @@
    <markus@oberhumer.com>               <ezerotven+github@gmail.com>
  */
 
-#include "../headers.h"
+#include "system_headers.h"
 #include <algorithm>
 #define ACC_WANT_ACC_INCI_H 1
-#include "../miniacc.h"
-#define ACC_WANT_ACCLIB_GETOPT 1
-#define ACC_WANT_ACCLIB_HSREAD 1
-#define ACC_WANT_ACCLIB_MISC 1
-#define ACC_WANT_ACCLIB_VGET 1
+#include "miniacc.h"
+#define ACC_WANT_ACCLIB_GETOPT   1
+#define ACC_WANT_ACCLIB_HSREAD   1
+#define ACC_WANT_ACCLIB_MISC     1
+#define ACC_WANT_ACCLIB_VGET     1
 #define ACC_WANT_ACCLIB_WILDARGV 1
 #undef HAVE_MKDIR
-#include "../miniacc.h"
+#include "miniacc.h"
 #include "../conf.h"
 
 /*************************************************************************
-// assert sane memory buffer sizes to protect against integer overflows
-// and malicious header fields
+// upx_rsize_t and mem_size: assert sane memory buffer sizes to protect
+// against integer overflows and malicious header fields
 // see C 11 standard, Annex K
 **************************************************************************/
 
+// this limits uncompressed_size to about 682 MiB (715_128_832 bytes)
 ACC_COMPILE_TIME_ASSERT_HEADER(UPX_RSIZE_MAX_MEM == UPX_RSIZE_MAX)
 ACC_COMPILE_TIME_ASSERT_HEADER(UPX_RSIZE_MAX_STR <= UPX_RSIZE_MAX / 256)
 ACC_COMPILE_TIME_ASSERT_HEADER(2ull * UPX_RSIZE_MAX * 9 / 8 + 256 * 1024 * 1024 < INT_MAX)
 ACC_COMPILE_TIME_ASSERT_HEADER(2ull * UPX_RSIZE_MAX * 10 / 8 + 128 * 1024 * 1024 <= INT_MAX + 1u)
-ACC_COMPILE_TIME_ASSERT_HEADER(5ull * UPX_RSIZE_MAX < UINT_MAX)
+ACC_COMPILE_TIME_ASSERT_HEADER(5ull * UPX_RSIZE_MAX < UINT_MAX) // IMPORTANT overflow protection
 ACC_COMPILE_TIME_ASSERT_HEADER(UPX_RSIZE_MAX >= 8192 * 65536)
 ACC_COMPILE_TIME_ASSERT_HEADER(UPX_RSIZE_MAX_STR >= 1024)
 
-upx_rsize_t mem_size(upx_uint64_t element_size, upx_uint64_t n, upx_uint64_t extra1,
-                     upx_uint64_t extra2) {
-    assert(element_size > 0);
-    if very_unlikely (element_size == 0 || element_size > UPX_RSIZE_MAX)
-        throwCantPack("mem_size 1; take care");
-    if very_unlikely (n > UPX_RSIZE_MAX)
-        throwCantPack("mem_size 2; take care");
-    if very_unlikely (extra1 > UPX_RSIZE_MAX)
-        throwCantPack("mem_size 3; take care");
-    if very_unlikely (extra2 > UPX_RSIZE_MAX)
-        throwCantPack("mem_size 4; take care");
-    upx_uint64_t bytes = element_size * n + extra1 + extra2; // cannot overflow
-    if very_unlikely (bytes > UPX_RSIZE_MAX)
-        throwCantPack("mem_size 5; take care");
-    return ACC_ICONV(upx_rsize_t, bytes);
-}
-
 bool mem_size_valid(upx_uint64_t element_size, upx_uint64_t n, upx_uint64_t extra1,
                     upx_uint64_t extra2) noexcept {
     assert_noexcept(element_size > 0);
@@ -86,6 +70,23 @@ bool mem_size_valid(upx_uint64_t element_size, upx_uint64_t n, upx_uint64_t extr
     return true;
 }
 
+upx_rsize_t mem_size(upx_uint64_t element_size, upx_uint64_t n, upx_uint64_t extra1,
+                     upx_uint64_t extra2) {
+    assert(element_size > 0);
+    if very_unlikely (element_size == 0 || element_size > UPX_RSIZE_MAX)
+        throwCantPack("mem_size 1; take care");
+    if very_unlikely (n > UPX_RSIZE_MAX)
+        throwCantPack("mem_size 2; take care");
+    if very_unlikely (extra1 > UPX_RSIZE_MAX)
+        throwCantPack("mem_size 3; take care");
+    if very_unlikely (extra2 > UPX_RSIZE_MAX)
+        throwCantPack("mem_size 4; take care");
+    upx_uint64_t bytes = element_size * n + extra1 + extra2; // cannot overflow
+    if very_unlikely (bytes > UPX_RSIZE_MAX)
+        throwCantPack("mem_size 5; take care");
+    return ACC_ICONV(upx_rsize_t, bytes);
+}
+
 TEST_CASE("mem_size") {
     CHECK(mem_size_valid(1, 0));
     CHECK(mem_size_valid(1, 0x30000000));
@@ -140,6 +141,7 @@ TEST_CASE("ptr_diff") {
     CHECK(ptr_udiff(buf, buf) == 0);
     CHECK(ptr_udiff(buf + 1, buf) == 1);
     CHECK_THROWS(ptr_udiff(buf, buf + 1));
+    UNUSED(buf);
 }
 
 // check that 2 buffers do not overlap; will throw on error
@@ -150,8 +152,9 @@ void uintptr_check_no_overlap(upx_uintptr_t a, size_t a_size, upx_uintptr_t b, s
     upx_uintptr_t b_end = b + mem_size(1, b_size);
     if very_unlikely (a_end < a || b_end < b) // wrap-around
         throwCantPack("ptr_check_no_overlap-overflow");
-    // same as (!(a >= b_end || b >= a_end))
-    // same as (!(a_end <= b || b_end <= a))
+    // simple, but a little bit mind bending:
+    //   same as (!(a >= b_end || b >= a_end))
+    //   same as (!(a_end <= b || b_end <= a))
     if very_unlikely (a < b_end && b < a_end)
         throwCantPack("ptr_check_no_overlap-ab");
 }
@@ -174,7 +177,7 @@ void uintptr_check_no_overlap(upx_uintptr_t a, size_t a_size, upx_uintptr_t b, s
         throwCantPack("ptr_check_no_overlap-bc");
 }
 
-#if DEBUG
+#if !defined(DOCTEST_CONFIG_DISABLE) && !defined(__wasi__) && DEBUG
 TEST_CASE("ptr_check_no_overlap 2") {
     byte p[4] = {};
 
@@ -251,7 +254,7 @@ TEST_CASE("ptr_check_no_overlap 3") {
 // stdlib
 **************************************************************************/
 
-void *upx_calloc(size_t n, size_t element_size) {
+void *upx_calloc(size_t n, size_t element_size) may_throw {
     size_t bytes = mem_size(element_size, n); // assert size
     void *p = malloc(bytes);
     if (p != nullptr)
@@ -260,85 +263,217 @@ void *upx_calloc(size_t n, size_t element_size) {
 }
 
 // simple unoptimized memswap()
-void upx_memswap(void *a, void *b, size_t n) {
+void upx_memswap(void *a, void *b, size_t n) noexcept {
     if (a != b && n != 0) {
-        char *x = (char *) a;
-        char *y = (char *) b;
+        byte *x = (byte *) a;
+        byte *y = (byte *) b;
         do {
             // strange clang-analyzer-15 false positive when compiling in Debug mode
             // clang-analyzer-core.uninitialized.Assign
-            char tmp = *x; // NOLINT(*core.uninitialized.Assign) // bogus clang-analyzer warning
+            byte tmp = *x; // NOLINT(*core.uninitialized.Assign) // bogus clang-analyzer warning
             *x++ = *y;
             *y++ = tmp;
         } while (--n != 0);
     }
 }
 
+// much better memswap(), optimized for our use case in sort functions below
+static void memswap_no_overlap(byte *a, byte *b, size_t n) noexcept {
+#if defined(__clang__) && __clang_major__ < 15
+    // work around a clang < 15 ICE (Internal Compiler Error)
+    // @COMPILER_BUG @CLANG_BUG
+    upx_memswap(a, b, n);
+#else // clang bug
+    upx_alignas_max byte tmp_buf[16];
+#define SWAP(x)                                                                                    \
+    do {                                                                                           \
+        upx_memcpy_inline(tmp_buf, a, x);                                                          \
+        upx_memcpy_inline(a, b, x);                                                                \
+        upx_memcpy_inline(b, tmp_buf, x);                                                          \
+        a += x;                                                                                    \
+        b += x;                                                                                    \
+    } while (0)
+
+    for (; n >= 16; n -= 16)
+        SWAP(16);
+    if (n & 8)
+        SWAP(8);
+    if (n & 4)
+        SWAP(4);
+    if (n & 2)
+        SWAP(2);
+    if (n & 1) {
+        byte tmp = *a;
+        *a = *b;
+        *b = tmp;
+    }
+#undef SWAP
+#endif // clang bug
+}
+
 // extremely simple (and beautiful) stable sort: Gnomesort
 // WARNING: O(n^2) and thus very inefficient for large n
-void upx_stable_sort(void *array, size_t n, size_t element_size,
-                     int (*compare)(const void *, const void *)) {
+void upx_gnomesort(void *array, size_t n, size_t element_size, upx_compare_func_t compare) {
     for (size_t i = 1; i < n; i++) {
-        char *a = (char *) array + element_size * i;        // a := &array[i]
-        if (i != 0 && compare(a - element_size, a) > 0) {   // if a[-1] > a[0] then
-            upx_memswap(a - element_size, a, element_size); //   swap elements a[-1] <=> a[0]
-            i -= 2;                                         //   and decrease i
+        byte *a = (byte *) array + element_size * i;               // a := &array[i]
+        if (i != 0 && compare(a - element_size, a) > 0) {          // if a[-1] > a[0] then
+            memswap_no_overlap(a - element_size, a, element_size); //   swap elements a[-1] <=> a[0]
+            i -= 2;                                                //   and decrease i
         }
     }
 }
 
-#if DEBUG
-TEST_CASE("upx_stable_sort") {
-    {
-        unsigned a[] = {0, 1};
-        upx_stable_sort(a, 2, sizeof(*a), ne32_compare);
-        CHECK((a[0] == 0 && a[1] == 1));
-    }
-    {
-        unsigned a[] = {1, 0};
-        upx_stable_sort(a, 2, sizeof(*a), ne32_compare);
-        CHECK((a[0] == 0 && a[1] == 1));
-    }
-    {
-        BE64 a[3];
-        a[0] = 257;
-        a[1] = 256;
-        a[2] = 255;
-        upx_stable_sort(a, 3, sizeof(*a), be64_compare);
-        CHECK((a[0] == 255 && a[1] == 256 && a[2] == 257));
+// simple Shell sort using Knuth's gap; NOT stable; uses memswap()
+// cannot compete with modern sort algorithms, but not too bad as a generic fallback
+void upx_shellsort_memswap(void *array, size_t n, size_t element_size, upx_compare_func_t compare) {
+    mem_size_assert(element_size, n); // check size
+    size_t gap = 0;                   // 0, 1, 4, 13, 40, 121, 364, 1093, ...
+    while (gap * 3 + 1 < n)           // cannot overflow because of size check above
+        gap = gap * 3 + 1;
+    for (; gap > 0; gap = (gap - 1) / 3) {
+        const size_t gap_bytes = element_size * gap;
+        byte *p = (byte *) array + gap_bytes;
+        for (size_t i = gap; i < n; i += gap, p += gap_bytes) // invariant: p == &array[i]
+            for (byte *a = p; a != array && compare(a - gap_bytes, a) > 0; a -= gap_bytes)
+                memswap_no_overlap(a - gap_bytes, a, element_size);
     }
 }
 
+// simple Shell sort using Knuth's gap; NOT stable; uses memcpy()
+// should be faster than memswap() version in theory, but benchmarks are inconsistent
+void upx_shellsort_memcpy(void *array, size_t n, size_t element_size, upx_compare_func_t compare) {
+    mem_size_assert(element_size, n); // check size
+    constexpr size_t MAX_INLINE_ELEMENT_SIZE = 256;
+    upx_alignas_max byte tmp_buf[MAX_INLINE_ELEMENT_SIZE]; // buffer for one element
+    byte *tmp = tmp_buf;
+    if (element_size > MAX_INLINE_ELEMENT_SIZE) {
+        tmp = (byte *) malloc(element_size);
+        assert(tmp != nullptr);
+    }
+    size_t gap = 0;         // 0, 1, 4, 13, 40, 121, 364, 1093, ...
+    while (gap * 3 + 1 < n) // cannot overflow because of size check above
+        gap = gap * 3 + 1;
+    for (; gap > 0; gap = (gap - 1) / 3) {
+        const size_t gap_bytes = element_size * gap;
+        byte *p = (byte *) array + gap_bytes;
+        for (size_t i = gap; i < n; i += gap, p += gap_bytes) // invariant: p == &array[i]
+            if (compare(p - gap_bytes, p) > 0) {
+                byte *a = p;
+                memcpy(tmp, a, element_size);
+                do {
+                    memcpy(a, a - gap_bytes, element_size);
+                    a -= gap_bytes;
+                } while (a != array && compare(a - gap_bytes, tmp) > 0);
+                memcpy(a, tmp, element_size);
+            }
+    }
+    if (element_size > MAX_INLINE_ELEMENT_SIZE)
+        free(tmp);
+}
+
+// wrap std::stable_sort()
+template <size_t ElementSize>
+void upx_std_stable_sort(void *array, size_t n, upx_compare_func_t compare) {
+    static_assert(ElementSize > 0 && ElementSize <= UPX_RSIZE_MAX);
+    mem_size_assert(ElementSize, n); // check size
+#if 0
+    // just for testing
+    upx_gnomesort(array, n, ElementSize, compare);
+#else
+    struct alignas(1) element_type { byte data[ElementSize]; };
+    static_assert(sizeof(element_type) == ElementSize);
+    static_assert(alignof(element_type) == 1);
+    auto less = [compare](const element_type &a, const element_type &b) -> bool {
+        return compare(&a, &b) < 0;
+    };
+    std::stable_sort((element_type *) array, (element_type *) array + n, less);
+#endif
+}
+
+#if UPX_CONFIG_USE_STABLE_SORT
+// instantiate function templates for all element sizes we need; efficient run time, but code size
+// bloat (about 4KiB code size for each function with my current libstdc++); not really needed as
+// libc qsort() is good enough for our use cases
+template void upx_std_stable_sort<1>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<2>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<4>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<5>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<8>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<16>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<32>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<56>(void *, size_t, upx_compare_func_t);
+template void upx_std_stable_sort<72>(void *, size_t, upx_compare_func_t);
+#endif
+
+#if !defined(DOCTEST_CONFIG_DISABLE) && DEBUG >= 1
 #if __cplusplus >= 202002L // use C++20 std::next_permutation() to test all permutations
 namespace {
+template <class ElementType, upx_compare_func_t CompareFunc>
 struct TestSortAllPermutations {
-    static upx_uint64_t test(size_t n) {
+    typedef ElementType element_type;
+    static noinline upx_uint64_t test(upx_sort_func_t sort, size_t n) {
         constexpr size_t N = 16;
-        assert(n > 0 && n <= N);
-        LE16 perm[N];
+        assert_noexcept(n <= N);
+        ElementType perm[N];
+        if (n == 0) {
+            sort(perm, 0, sizeof(ElementType), CompareFunc); // check that n == 0 works
+            return 0;
+        }
         for (size_t i = 0; i < n; i++)
             perm[i] = 255 + i;
         upx_uint64_t num_perms = 0;
         do {
-            LE16 a[N];
+            ElementType a[N];
             memcpy(a, perm, sizeof(*a) * n);
-            upx_stable_sort(a, n, sizeof(*a), le16_compare);
+            sort(a, n, sizeof(*a), CompareFunc);
             for (size_t i = 0; i < n; i++)
-                assert((a[i] == 255 + i));
+                assert_noexcept((a[i] == 255 + i));
             num_perms += 1;
         } while (std::next_permutation(perm, perm + n));
         return num_perms;
     }
+    static noinline bool test_permutations(upx_sort_func_t sort) {
+        bool ok = true;
+        ok &= (test(sort, 0) == 0);
+        ok &= (test(sort, 1) == 1);
+        ok &= (test(sort, 2) == 2);
+        ok &= (test(sort, 3) == 6);
+        ok &= (test(sort, 4) == 24);
+        ok &= (test(sort, 5) == 120);
+#if DEBUG >= 2
+        ok &= (test(sort, 6) == 720);
+        ok &= (test(sort, 7) == 5040);
+        ok &= (test(sort, 8) == 40320);
+        ok &= (test(sort, 9) == 362880);
+        ok &= (test(sort, 10) == 3628800);
+        // ok &= (test(sort, 11) == 39916800);
+#endif
+        return ok;
+    }
 };
 } // namespace
-TEST_CASE("upx_stable_sort") {
-    CHECK(TestSortAllPermutations::test(1) == 1);
-    CHECK(TestSortAllPermutations::test(2) == 2);
-    CHECK(TestSortAllPermutations::test(3) == 6);
-    CHECK(TestSortAllPermutations::test(4) == 24);
-    CHECK(TestSortAllPermutations::test(5) == 120);
-    // CHECK(TestSortAllPermutations::test(6) == 720);
-    // CHECK(TestSortAllPermutations::test(7) == 5040);
+TEST_CASE("upx_gnomesort") {
+    // typedef TestSortAllPermutations<BE64, be64_compare> TestSort;
+    typedef TestSortAllPermutations<LE16, le16_compare> TestSort;
+    CHECK(TestSort::test_permutations(upx_gnomesort));
+}
+TEST_CASE("upx_shellsort_memswap") {
+    // typedef TestSortAllPermutations<BE64, be64_compare> TestSort;
+    typedef TestSortAllPermutations<LE16, le16_compare> TestSort;
+    CHECK(TestSort::test_permutations(upx_shellsort_memswap));
+}
+TEST_CASE("upx_shellsort_memcpy") {
+    // typedef TestSortAllPermutations<BE64, be64_compare> TestSort;
+    typedef TestSortAllPermutations<LE16, le16_compare> TestSort;
+    CHECK(TestSort::test_permutations(upx_shellsort_memcpy));
+}
+TEST_CASE("upx_std_stable_sort") {
+    // typedef TestSortAllPermutations<BE64, be64_compare> TestSort;
+    typedef TestSortAllPermutations<LE16, le16_compare> TestSort;
+    upx_sort_func_t wrap_stable_sort = [](void *a, size_t n, size_t, upx_compare_func_t compare) {
+        upx_std_stable_sort<sizeof(TestSort::element_type)>(a, n, compare);
+    };
+    CHECK(TestSort::test_permutations(wrap_stable_sort));
 }
 #endif // C++20
 #endif // DEBUG
@@ -518,12 +653,13 @@ TEST_CASE("find") {
     CHECK(find_le64(b, 16, 0x0f0e0d0c0b0a0908ULL) == 8);
     CHECK(find_be64(b, 15, 0x08090a0b0c0d0e0fULL) == -1);
     CHECK(find_le64(b, 15, 0x0f0e0d0c0b0a0908ULL) == -1);
+    UNUSED(b);
 }
 
 int mem_replace(void *buf, int blen, const void *what, int wlen, const void *replacement) noexcept {
-    byte *b = (byte *) buf;
+    byte *const b = (byte *) buf;
     int boff = 0;
-    int n = 0;
+    int count = 0;
 
     while (blen - boff >= wlen) {
         int off = find(b + boff, blen - boff, what, wlen);
@@ -532,9 +668,9 @@ int mem_replace(void *buf, int blen, const void *what, int wlen, const void *rep
         boff += off;
         memcpy(b + boff, replacement, wlen);
         boff += wlen;
-        n++;
+        count++;
     }
-    return n;
+    return count;
 }
 
 TEST_CASE("mem_replace") {
@@ -545,6 +681,7 @@ TEST_CASE("mem_replace") {
     CHECK(mem_replace(b + 8, 8, "bbb", 3, "efg") == 2);
     CHECK(mem_replace(b, 16, "b", 1, "h") == 2);
     CHECK(strcmp(b, "cdcdcdcdefgefghh") == 0);
+    UNUSED(b);
 }
 
 /*************************************************************************
@@ -569,18 +706,18 @@ const LEPolicy le_policy;
      ACC_OS_TOS || ACC_OS_WIN16 || ACC_OS_WIN32 || ACC_OS_WIN64)
 
 static const char dir_sep[] = "/\\";
-#define fn_is_drive(s) (s[0] && s[1] == ':')
-#define fn_is_sep(c) (strchr(dir_sep, c) != nullptr)
+#define fn_is_drive(s)   (s[0] && s[1] == ':')
+#define fn_is_sep(c)     (strchr(dir_sep, c) != nullptr)
 #define fn_skip_drive(s) (fn_is_drive(s) ? (s) + 2 : (s))
-#define fn_tolower(c) (tolower(((uchar) (c))))
+#define fn_tolower(c)    (tolower(((uchar) (c))))
 
 #else
 
 // static const char dir_sep[] = "/";
-#define fn_is_drive(s) (0)
-#define fn_is_sep(c) ((c) == '/')
+#define fn_is_drive(s)   (0)
+#define fn_is_sep(c)     ((c) == '/')
 #define fn_skip_drive(s) (s)
-#define fn_tolower(c) (c)
+#define fn_tolower(c)    (c)
 
 #endif
 
@@ -629,9 +766,22 @@ int fn_strcmp(const char *n1, const char *n2) {
 }
 
 /*************************************************************************
-// misc.
+// misc
 **************************************************************************/
 
+bool is_envvar_true(const char *envvar, const char *alternate_name) noexcept {
+    // UPX convention: any environment variable that is set and is not strictly equal to "0" is true
+    const char *e = getenv(envvar);
+    if (e != nullptr && e[0])
+        return strcmp(e, "0") != 0;
+    if (alternate_name != nullptr) {
+        e = getenv(alternate_name);
+        if (e != nullptr && e[0])
+            return strcmp(e, "0") != 0;
+    }
+    return false;
+}
+
 bool set_method_name(char *buf, size_t size, int method, int level) {
     bool r = true;
     const char *alg;
@@ -787,4 +937,15 @@ TEST_CASE("get_ratio") {
     CHECK(get_ratio(2 * UPX_RSIZE_MAX, 1024ull * UPX_RSIZE_MAX) == 9999999);
 }
 
+/*************************************************************************
+// compat
+**************************************************************************/
+
+#if defined(__wasi__) // TODO later - wait for wasm/wasi exception handling proposal
+extern "C" {
+void __cxa_allocate_exception() { std::terminate(); }
+void __cxa_throw() { std::terminate(); }
+} // extern "C"
+#endif
+
 /* vim:set ts=4 sw=4 et: */
